// ======================================================================
//
// MaterialData_3dfx.cpp
// jeff grills
//
// copyright 1996 origin systems inc
//
// use a pointer to a member function to set up hardware state to avoid
// doing conditional branches.
//
// ======================================================================

#include "first.h"
#include "MaterialData_3dfx.h"

#include "Gl_3dfx.h"
#include "Material.h"
#include "PaletteData_3dfx.h"
#include "Palette.h"
#include "TextureManager_3dfx.h"
#include "uvstruct.h"

// ======================================================================

u_int MaterialData_3dfx::globalTagLRU;

// buffer to build converted textures
static union
{
	Pixel16 pixel16[256 * 256 + 128 * 128 + 64 * 64 + 32 * 32 + 16 * 16 + 8 * 8 + 4 * 4 + 2 * 2 + 1 * 1];
	Pixel8  pixel8 [256 * 256 + 128 * 128 + 64 * 64 + 32 * 32 + 16 * 16 + 8 * 8 + 4 * 4 + 2 * 2 + 1 * 1];
};

// ======================================================================

static GrLOD_t computeLOD(int width)
{
	switch (width)
	{
		case 256: return GR_LOD_256;
		case 128: return GR_LOD_128;
		case  64: return GR_LOD_64; 
		case  32: return GR_LOD_32; 
		case  16: return GR_LOD_16; 
		case   8: return GR_LOD_8;  
		case   4: return GR_LOD_4;  
		case   2: return GR_LOD_2;  
		case   1: return GR_LOD_1;  
	}

	DEBUG_FATAL(true, ("Mat_3dfx::computeLOD bad width %dx", width));
	return GR_LOD_1;
}

// ======================================================================

MaterialData_3dfx::MaterialData_3dfx(const StaticMaterial *material)
{
	init();

	// remember our creator material
	staticMaterial = material;
	
	const Pixel8 * const rgb   = material->getRGBMap8();
	const Pixel8 * const alpha = material->getAlphaMap();

	if (rgb == NULL && alpha == NULL)
	{
		// solid material
		color = (material->getA() << 24) | (material->getR() << 16) | (material->getG() << 8) | (material->getB() << 0);

		if (material->getA() == 255)
		{
			translucency  = false;
			setupFunction = setupChromakeyOffColorConstantAlphaOff;
		}
		else
		{
			translucency  = true;
			setupFunction = setupChromakeyOffColorConstantAlphaConstant;
		}

		return;
	}

	const int         width = material->getWidth();
	const int         height = material->getHeight();
	GrTextureFormat_t format;
	GrLOD_t           smallLOD, largeLOD;

	largeLOD = computeLOD(width);
	smallLOD = computeLOD(width >> (material->getMipMapLevels() - 1));

	bool solidWithAlpha = (rgb == NULL);
	bool translucent    = material->getAlpha() != StaticMaterial::AlphaOff;
	bool transparent    = material->getChromakey() != StaticMaterial::ChromakeyOff;

	if (solidWithAlpha)
	{
		translucency  = true;
		setupFunction = setupChromakeyOffColorConstantAlphaMap;
		format = GR_TEXFMT_ALPHA_8;
	}
	else
		if (translucent)
		{
			translucency  = true;
			setupFunction  = setupChromakeyOffColorMapAlphaMap;
			format = GR_TEXFMT_ARGB_4444;
		}
		else
			if (transparent)
			{
				translucency  = false;
				setupFunction  = setupChromakeyOnColorMapAlphaOff;
				format = GR_TEXFMT_ARGB_1555;
			}
			else
			{
				translucency  = false;
				setupFunction  = setupChromakeyOffColorMapAlphaOff;
				format = GR_TEXFMT_P_8;
			}

	// calculate the proper amount of texture memory
	size = grTexCalcMemRequired(smallLOD, largeLOD, GR_ASPECT_1x1, format);

	// setup the texture information
	info.smallLod    = smallLOD;
	info.largeLod    = largeLOD;
	info.aspectRatio = GR_ASPECT_1x1;
	info.format      = format;
	info.data        = pixel16;

	// set the solid color for solid with alpha
	if (format == GR_TEXFMT_ALPHA_8)
		color = ((material->getR() << 16) | (material->getG() << 8) | (material->getB() << 0));

	// load the palette data for the material
	if (format == GR_TEXFMT_P_8)
		palette = static_cast<PaletteData_3dfx *>(staticMaterial->getPalette()->getData());
}

// ----------------------------------------------------------------------

MaterialData_3dfx::MaterialData_3dfx(const uv_instance *uv)
{
	init();

	UNREF(uv);

	// size of space to allocate on the card (should be 256*256*2)
	size = grTexCalcMemRequired(GR_LOD_256, GR_LOD_256, GR_ASPECT_1x1, GR_TEXFMT_RGB_565);  

	// allocate texture ram for this texure
	ptr = TextureManager::alloc(size, this);

	info.smallLod    = GR_LOD_256;
	info.largeLod    = GR_LOD_256;
	info.aspectRatio = GR_ASPECT_1x1;
	info.format      = GR_TEXFMT_RGB_565;
	info.data        = pixel16;

	loaded = true;
}

// ----------------------------------------------------------------------

MaterialData_3dfx::~MaterialData_3dfx(void)
{
	unload();
	init();
}

// ----------------------------------------------------------------------

void MaterialData_3dfx::init(void)
{
	staticMaterial    = NULL;
	palette           = NULL;

	color             = 0;
	tagLRU            = 0;
	ptr               = 0;
	size              = 0;
	memset(&info, 0, sizeof(info));
	translucency      = false;
	loaded            = false;

	setupFunction     = setupNOP;
}

// ----------------------------------------------------------------------

void MaterialData_3dfx::download(void) const
{
	GrTexInfo info2;
	int       i, j, k;
	Pixel32   p;

	if (palette)
		palette->download();

	if (loaded)
		return;

	const int   levels       = staticMaterial->getMipMapLevels();
	int         width        = staticMaterial->getWidth();
	int         height       = staticMaterial->getHeight();
	const RGB  *pal          = (staticMaterial->getRGBMap8() ? staticMaterial->getPalette()->getRGB() : NULL);
	bool        transparency = staticMaterial->getChromakey() != StaticMaterial::ChromakeyOff;

	// allocate texture ram for this texure
	ptr = TextureManager::alloc(size, this);

	info2 = info;

	switch (info2.format)
	{
		case GR_TEXFMT_ALPHA_8:

			for (k = 0, j = 0; j < levels; ++j)
			{
				const Pixel8 *alpha = staticMaterial->getAlphaMap(j);
				const int size      = width * height;

				// convert to 3dfx-style alpha
				for (i = 0; i < size; ++i, ++k)
					pixel8[k] = static_cast<Pixel8>(alpha[i]);


				// reduce the width and height for the next mip map level
				width >>= 1;
				height >>= 1;
			}

			info2.data = pixel8;
			break;

		case GR_TEXFMT_ARGB_4444:

			for (k = 0, j = 0; j < levels; ++j)
			{
				const Pixel8 *rgb   = staticMaterial->getRGBMap8(j);
				const Pixel8 *alpha = staticMaterial->getAlphaMap(j);
				const int size      = width * height;

				// check for alpha map or constant alpha value
				if (alpha)
				{
					// convert to argb 4444
					for (i = 0; i < size; ++i, ++k)
					{
						p = rgb[i];
						if (p == 0 && transparency)
							pixel16[k] = 0x0000;
						else
							pixel16[k] = static_cast<Pixel16>(((alpha[i] >> 4) << 12) | ((pal[p].r >> 4) << 8) | ((pal[p].g >> 4) <<  4) | ((pal[p].b >> 4) <<  0));
					}
				}
				else
				{
					int a;

					a = (staticMaterial->getA() >> 4) << 12;

					// convert to argb 4444
					for (i = 0; i < size; ++i, ++k)
					{
						p = rgb[i];
						if (p == 0 && transparency)
							pixel16[k] = 0x0000;
						else
							pixel16[k] = static_cast<Pixel16>(a | ((pal[p].r >> 4) << 8) | ((pal[p].g >> 4) <<  4) | ((pal[p].b >> 4) <<  0));
					}
				}

				// reduce the width and height for the next mip map level
				width >>= 1;
				height >>= 1;
			}

			break;

		case GR_TEXFMT_ARGB_1555:

			for (k = 0, j = 0; j < levels; ++j)
			{
				const Pixel8 *rgb = staticMaterial->getRGBMap8(j);
				const int size    = width * height;

				for (i = 0; i < size; ++i, ++k)
				{
					p = rgb[i];
					if (p == 0 && transparency)
						pixel16[k] = 0x0000;
					else
						pixel16[k] = static_cast<Pixel16>(0x8000 | ((pal[p].r >> 3) << 10) | ((pal[p].g >> 3) <<  5) | ((pal[p].b >> 3) <<  0));
				}

				// reduce the width and height for the next mip map level
				width >>= 1;
				height >>= 1;
			}

			break;

		case GR_TEXFMT_P_8:

			for (k = 0, j = 0; j < levels; ++j)
			{
				const Pixel8 *rgb = staticMaterial->getRGBMap8(j);
				const int size    = width * height;

				for (i = 0; i < size; ++i, ++k)
					pixel8[k] = rgb[i];

				// reduce the width and height for the next mip map level
				width >>= 1;
				height >>= 1;
			}

			info2.data = pixel8;
			break;

		default:
			DEBUG_FATAL(true, ("MaterialData_3dfx::download bad format\n"));
			return;
	}

	// download the image into the texture space
	grTexDownloadMipMap(GR_TMU0, ptr, GR_MIPMAPLEVELMASK_BOTH, &info2);
	GL_3dfx::countDownloadedTexture();
	loaded = true;
}

// ----------------------------------------------------------------------

void MaterialData_3dfx::download(const uv_instance *uv,u_int32 squareXOffset,u_int32 squareYOffset) const
{
	GrTexInfo info2;

	u_int32 i;
	u_int32 quickWidth=uv->width;
	u_int32 quickHeight=uv->height;

	if (uv->options.option_flags & (X_DOUBLE | X_SKIP))  // adjust from # of pixels to actual size of the buffer
		quickWidth*=2;
	if (uv->options.option_flags & (Y_DOUBLE | Y_SKIP))
		quickHeight*=2;
	{
		u_int32 cappedX,cappedY;
		u_int32 srcIndex,destIndex;

		cappedX=min(squareXOffset+256,quickWidth);  // don't extend beyond the edges of the movie source
		cappedY=min(squareYOffset+256,quickHeight);

		destIndex=0;

		if (uv->options.option_flags & Y_SKIP)  // if Y is skipped
		{
			srcIndex=(((quickWidth*squareYOffset)>>1)+squareXOffset)*2;  // double the source index to save a multiply/shift in the memcpy

			for (i = squareYOffset; i < cappedY; i+=2, destIndex+=512, srcIndex+=quickWidth*2)
			{
				memcpy(pixel16+destIndex,uv->last_buffer+srcIndex,(cappedX-squareXOffset)*sizeof(Pixel16));
				memset(pixel16+destIndex+256,0,(cappedX-squareXOffset)*sizeof(Pixel16));
			}
		}
		else if (uv->options.option_flags & Y_DOUBLE)  // if Y is doubled, don't index further into the source, cuz it isn't there
		{
			srcIndex=((quickWidth*squareYOffset)>>1)+squareXOffset;

			for (i = squareYOffset; i < cappedY; i+=2, destIndex+=512, srcIndex+=quickWidth)
			{
				memcpy(pixel16+destIndex,uv->last_buffer+srcIndex+srcIndex,(cappedX-squareXOffset)*sizeof(Pixel16));
				memcpy(pixel16+destIndex+256,uv->last_buffer+srcIndex+srcIndex,(cappedX-squareXOffset)*sizeof(Pixel16));
			}
		}
		else 
		{
			srcIndex=quickWidth*squareYOffset+squareXOffset;
			for (i = squareYOffset; i < cappedY; ++i)  // no Y doubling or skipping
			{

				memcpy(pixel16+destIndex,uv->last_buffer+srcIndex+srcIndex,(cappedX-squareXOffset)*sizeof(Pixel16));
				destIndex+=256;
				srcIndex+=quickWidth;
			}
		}
	}

	// download the image into the texture space
	info2=info;
	grTexDownloadMipMap(GR_TMU0, ptr, GR_MIPMAPLEVELMASK_BOTH, &info2);
	GL_3dfx::countDownloadedTexture();
}

// ----------------------------------------------------------------------

void MaterialData_3dfx::unload(void) const
{
	if (loaded)
	{
		TextureManager::free(ptr);
		ptr = 0;
		loaded = false;
	}
}

// ----------------------------------------------------------------------

void MaterialData_3dfx::setupNOP(void) const
{
}

// ----------------------------------------------------------------------

void MaterialData_3dfx::setupChromakeyOffColorConstantAlphaOff(void) const
{
	grChromakeyMode(GR_CHROMAKEY_DISABLE);

	grConstantColorValue(color);

	grAlphaBlendFunction(GR_BLEND_ONE, GR_BLEND_ZERO, GR_BLEND_ONE, GR_BLEND_ZERO);
	grAlphaCombine(GR_COMBINE_FUNCTION_BLEND_OTHER, GR_COMBINE_FACTOR_LOCAL, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_CONSTANT, FXFALSE);
	grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER, GR_COMBINE_FACTOR_LOCAL, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_CONSTANT, FXFALSE);
}

// ----------------------------------------------------------------------

void MaterialData_3dfx::setupChromakeyOffColorConstantAlphaConstant(void) const
{
	grChromakeyMode(GR_CHROMAKEY_DISABLE);

	grConstantColorValue(color);

	grAlphaBlendFunction(GR_BLEND_SRC_ALPHA, GR_BLEND_ONE_MINUS_SRC_ALPHA, GR_BLEND_ONE, GR_BLEND_ZERO);
	grAlphaCombine(GR_COMBINE_FUNCTION_BLEND_OTHER, GR_COMBINE_FACTOR_LOCAL, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_CONSTANT, FXFALSE);
	grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER, GR_COMBINE_FACTOR_LOCAL, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_CONSTANT, FXFALSE);
}

// ----------------------------------------------------------------------

void MaterialData_3dfx::setupChromakeyOffColorConstantAlphaMap(void) const
{
	download();

	grChromakeyMode(GR_CHROMAKEY_DISABLE);

	grConstantColorValue(color);

	grAlphaBlendFunction(GR_BLEND_SRC_ALPHA, GR_BLEND_ONE_MINUS_SRC_ALPHA, GR_BLEND_ONE, GR_BLEND_ZERO);
	grAlphaCombine(GR_COMBINE_FUNCTION_BLEND_OTHER, GR_COMBINE_FACTOR_LOCAL, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_TEXTURE, FXFALSE);
	grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER, GR_COMBINE_FACTOR_LOCAL, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_CONSTANT, FXFALSE);

	grTexSource(GR_TMU0, ptr, GR_MIPMAPLEVELMASK_BOTH, const_cast<GrTexInfo *>(&info));
}

// ----------------------------------------------------------------------

void MaterialData_3dfx::setupChromakeyOffColorMapAlphaOff(void) const
{
	download();

	grChromakeyMode(GR_CHROMAKEY_DISABLE);

	grConstantColorValue(color);

	grAlphaBlendFunction(GR_BLEND_ONE, GR_BLEND_ZERO, GR_BLEND_ONE, GR_BLEND_ZERO);
	grAlphaCombine(GR_COMBINE_FUNCTION_BLEND_OTHER, GR_COMBINE_FACTOR_LOCAL, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_CONSTANT, FXFALSE);
	grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER, GR_COMBINE_FACTOR_LOCAL, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_TEXTURE, FXFALSE);

	grTexSource(GR_TMU0, ptr, GR_MIPMAPLEVELMASK_BOTH, const_cast<GrTexInfo *>(&info));
}

// ----------------------------------------------------------------------

void MaterialData_3dfx::setupChromakeyOffColorMapAlphaMap(void) const
{
	download();

	grChromakeyMode(GR_CHROMAKEY_DISABLE);

	grConstantColorValue(color);

	grAlphaBlendFunction(GR_BLEND_SRC_ALPHA, GR_BLEND_ONE_MINUS_SRC_ALPHA, GR_BLEND_ONE, GR_BLEND_ZERO);
	grAlphaCombine(GR_COMBINE_FUNCTION_BLEND_OTHER, GR_COMBINE_FACTOR_LOCAL, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_TEXTURE, FXFALSE);
	grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER, GR_COMBINE_FACTOR_LOCAL, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_TEXTURE, FXFALSE);

	grTexSource(GR_TMU0, ptr, GR_MIPMAPLEVELMASK_BOTH, const_cast<GrTexInfo *>(&info));
}

// ----------------------------------------------------------------------

void MaterialData_3dfx::setupChromakeyOffColorMapAlphaConstant(void) const
{
	download();

	grChromakeyMode(GR_CHROMAKEY_DISABLE);

	grConstantColorValue(color);

	grAlphaBlendFunction(GR_BLEND_SRC_ALPHA, GR_BLEND_ONE_MINUS_SRC_ALPHA, GR_BLEND_ONE, GR_BLEND_ZERO);
	grAlphaCombine(GR_COMBINE_FUNCTION_BLEND_OTHER, GR_COMBINE_FACTOR_LOCAL, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_CONSTANT, FXFALSE);
	grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER, GR_COMBINE_FACTOR_LOCAL, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_TEXTURE, FXFALSE);

	grTexSource(GR_TMU0, ptr, GR_MIPMAPLEVELMASK_BOTH, const_cast<GrTexInfo *>(&info));
}

// ----------------------------------------------------------------------

void MaterialData_3dfx::setupChromakeyOnColorMapAlphaOff(void) const
{
	download();

	grChromakeyMode(GR_CHROMAKEY_ENABLE);

	grConstantColorValue(color);

	grAlphaBlendFunction(GR_BLEND_ONE, GR_BLEND_ZERO, GR_BLEND_ONE, GR_BLEND_ZERO);
	grAlphaCombine(GR_COMBINE_FUNCTION_BLEND_OTHER, GR_COMBINE_FACTOR_LOCAL, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_CONSTANT, FXFALSE);
	grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER, GR_COMBINE_FACTOR_LOCAL, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_TEXTURE, FXFALSE);

	grTexSource(GR_TMU0, ptr, GR_MIPMAPLEVELMASK_BOTH, const_cast<GrTexInfo *>(&info));
}

// ======================================================================
